/*
 .T ABCp A generic ABC Parser
 .A ...
 .d Aug 2004

 .abstract
   This is the '/scanner/ used in ABCp. The main function
   is '{abcNextToken()} that returns the next token in a file
   based on the current state.
 ..

 .disclaimer
   This code is distributed under the MIT licence (the same as Lua)
   The full text is at http://opensource.org/licenses/mit-license.php.
   The short story is that you are free to use this code as you wish
   as long as the appropriate credit to authors is given in a convenient
   place in your application and in the associated documentation.
   
   ABCp uses '|re2C|. See '<http://re2c.sourceforge.net> for details.
 ..
 
 .* How to read this file
 
 You should start from the abcNextFunction.
 
*/

#include "abcp.h"

void abcClose(abcScanner *s);
abcScanner *abcOpen(char *filename,USHORT bufsize);
int abcNextToken(abcScanner *s);

#ifdef DLL
#define ABCAPI __declspec(dllexport)

BOOL WINAPI __declspec(dllexport) LibMain(HINSTANCE hDLLInst, DWORD fdwReason, LPVOID lpvReserved)
{
    switch (fdwReason)
    {
        case DLL_PROCESS_ATTACH:
            break;
            
        case DLL_PROCESS_DETACH:
            break;
            
        case DLL_THREAD_ATTACH:
            break;
            
        case DLL_THREAD_DETACH:
            break;
    }
    return TRUE;
}


#else
#define ABCAPI
#endif

static char *abcTokName[]={
  "T_FIRST",
  "T_UNKNOWN",   "T_NONE",      "T_EOF",       "T_COMMENT",   "T_EMPTYLINE",
  "T_EXTFIELD",  "T_FIELD",     "T_NOTE",      "T_WSPACE",    "T_BAR",
  "T_INFIELD",   "T_TEXT",      "T_ENDLINE",   "T_CONTINUE",  "T_DECOR",
  "T_REST",      "T_INVREST",   "T_MULTIREST", "T_SPACER",    "T_NPLET",
  "T_OPENSLUR",  "T_CLOSESLUR", "T_DOTRIGHT",  "T_DOTLEFT",   "T_CHORDSTART",
  "T_CHORDEND",  "T_CHORD",     "T_REPEAT",    "T_GRACEAPP",  "T_GRACEACC",
  "T_GRACEEND",  "T_TIE",       "T_STRING",    "T_STRINGUNF", "T_ANNOTATION",
  "T_BREAK",     "T_OVERLAY",   "T_STARTLINE", "T_FIELDB",    "T_PRAGMA",
  "T_LYR_SYLL",  "T_LYR_BLANK", "T_LYR_BAR",   "T_LYR_SPACE", "T_LYR_CONT",
  "T_LYR_VERSE",
  "T_LAST"
};

static char *abcStName[]={
/*  0 */ "S_EOF", 
/*  1 */ "S_ERROR",
/*  2 */ "S_LIMBO",
/*  3 */ "S_NONE",
/*  4 */ "S_TUNE",
/*  5 */ "S_FIELD",
/*  6 */ "S_EXTFIELD",
/*  7 */ "S_INFIELD"
};

/*
** Line buffer limit
*/

#define MINBUFSIZE     1024
#define MAXBUFSIZE  5242880 
/*
 .* Macros
*/

/* To ease my paranoia about dynamic memory allocation
*/
#define m_guard(x)      (x)->guard=0xF00D
#define m_unguard(x)    (x)->guard=0x0BAD
#define m_guarded(x)    ((x) && ((x)->guard == 0xF00D))


/* Points to the last char in the available buffer space */
#define m_lastch(x)      (x)->buffer[(x)->bufsize-1]

/*                       if (*(x)->cursor == '\001') (x)->cursor++;\ */
 
/*
When a Match is found, the RETURN macro performs the following actions:

 1) The token field is set to the token 
 2) The "start of line char" ('\001') is skipped
 3) The length of token is computed
 4) The tokstr field is initialized to point to the start of the
    string that matched.
 5) The cursor is advanced right after the matched string
 6) The newline character, if present, is deleted
 7) The marker is set
 8) The first character of the next string to match is saved
    in the last char of the buffer.
 9) The first character of the next string is replaced by '\0' so
    that s->tokstr is a properly null terminated string.
    The saved char is to be restored next time the abcNextToken()
    is called.
10) The length of the string matched is returned.
    
*/
#define RETURN(x,t)   {(x)->token=t;\
                      if (*((x)->cursor) == '\001') (x)->cursor = (x)->cursor +1;\
                      (x)->toklen = YYCURSOR - (x)->cursor;\
                      (x)->tokstr = (x)->cursor;\
                      (x)->cursor = YYCURSOR;\
                      if (*((x)->cursor - 1) == '\n') *((x)->cursor - 1) = '\0';\
                      (x)->marker = marker;\
                      m_lastch(x) = *YYCURSOR;\
                      *((x)->cursor) = '\0';\
                      return (x)->toklen;}

#define m_linestart(x)  ((x)->cursor == (x)->buffer)
#define m_lineempty(x)  (*((x)->cursor) == '\0')
#define m_resetbuf(x)   {(x)->buffer[0]='\0';  \
                         (x)->buffer[1]='\0';  \
                         (x)->limit=(x)->buffer; \
                         (x)->cursor=(x)->limit; \
                         (x)->marker=(x)->cursor; \
                         (x)->tokstr=(x)->limit; \
                         m_lastch(x) ='\0'; \
                         (x)->toklen=0;}

#define m_eof(x)  {m_resetbuf(x); \
                   (x)->token=T_EOF;\
                   (x)->state=S_EOF;}

#define m_column(x) ((x)->tokstr - (x)->buffer)

/*
** The scanning routines
*/

static void abcGetLine(abcScanner *s)
{
  char *t=NULL;
  
  if (m_guarded(s) && (s->state != S_EOF)) {
    m_resetbuf(s);
    s->buffer[0]='\001';
    if (s->file) {
      t=fgets(s->buffer+1,s->bufsize-4,s->file);
      /*printf("*\t[%s]\n",s->buffer);*/

    }
    else if (*(s->bufaux))  {
      /*printf("+\t%s\n",s->bufaux);*/
      t=s->buffer+1;
      while (*(s->bufaux) && (*(s->bufaux) != '\n'))
        *t++ = *(s->bufaux)++;
      *t++ = *(s->bufaux);
      *t='\0';
      if (*(s->bufaux) != '\0') (s->bufaux)++;
    }
    if (t) {
      s->line++;
      s->cursor=s->buffer;
      while (*t) t++;
      s->limit=t;
      s->marker=s->cursor;
    }
    else
      m_eof(s);
  }
}


/*!re2c
any=[\002-\377];
EOS=[\000];
SOL=[\001];
NWS=[\041-\177];
WS=[\040\011];
NL=[\012\015]+;
L	= [A-Za-z];
D = [0-9];

BA1 = "|" | "[" | ":" ;
BA2 = "|" | "]" | ":" ;
BAR = ".|" | "|" | BA1 BA2+ ;

MIN =[\055];

ACC = "_" | "__" | "^" | "^^" | "=";
N=  ACC? [A-Ga-g] ("," | "'")* (D | "/")*;

DEC=[~.uvTHLMPSO] | "\053" (any\[\053\040\011])+ "\053" | "\041" (any\[\040\011\041])+ "\041" ;

Q=[\042];
NQ=any\Q;

ANN  =Q [@<>^_] NQ* Q;
CHRD =Q [A-G] NQ* Q;
STR  =Q NQ* Q;
BS   =[\134];

BRK  =[\041];
*/

static int abcNT(abcScanner *s)
{
  YYCTYPE *cursor, *limit, *marker;
  cursor = s->cursor;
  limit  = s->limit;
  marker = s->marker;
  
  /*!re2c

  EOS { return -1;}

  SOL WS* "%%" WS* {
      s->cursor=YYCURSOR;
      while (*YYCURSOR && !isspace(*YYCURSOR)) 
        YYCURSOR++;
      s->nextstate=S_EXTFIELD;
      RETURN(s,T_EXTFIELD);
  }

  SOL WS* L [:] {
      s->cursor=YYCURSOR-2;
      s->nextstate =  S_FIELD | *(s->cursor);
      RETURN(s,T_FIELD);
  }

  WS* [%]	{
      s->cursor=YYCURSOR;
      YYCURSOR=YYLIMIT;
      RETURN(s,T_COMMENT);
  }

  WS* [#]	{
      s->cursor=YYCURSOR;
      YYCURSOR=YYLIMIT;
      RETURN(s,T_PRAGMA);
  }

  SOL WS* NL {
      s->cursor++;
      s->nextstate=S_LIMBO;
      YYCURSOR=YYLIMIT;
      RETURN(s,T_EMPTYLINE);
  }
  
  BS WS* NL {
      YYCURSOR=YYLIMIT;
      RETURN(s,T_CONTINUE);
  }
    
  WS* NL {
      if (s->state & S_ANYFIELD) s->nextstate=S_TUNE;
      YYCURSOR=YYLIMIT;
      RETURN(s,T_ENDLINE);
  }

  [&]  { RETURN(s,T_OVERLAY); }
    
  any { return -1;}
  
  */
}

static int abcNT_FIELD_w(abcScanner *s)
{
  YYCTYPE *cursor, *limit, *marker;
  
  cursor = s->cursor;
  limit  = s->limit;
  marker = s->marker;
  
  /*!re2c
    SOL              { RETURN(s,T_STARTLINE); }
    
    WS+              { RETURN(s,T_LYR_SPACE); }

    (L | "\\" NWS | [(),.?!'~])+"-"?
                     { RETURN(s,T_LYR_SYLL); }
    
    [*]              { RETURN(s,T_LYR_BLANK); }

    MIN              { RETURN(s,T_LYR_SYLL); }
    
    [_]              { RETURN(s,T_LYR_CONT); }

    [|]              { RETURN(s,T_LYR_BAR); }
  
    any              { return -1; }
  */
  
}

static int abcNT_FIELD_s(abcScanner *s)
{
  YYCTYPE *cursor, *limit, *marker;
  
  cursor = s->cursor;
  limit  = s->limit;
  marker = s->marker;
  
  /*!re2c
    SOL                  { RETURN(s,T_STARTLINE); }
    
    WS+                  { RETURN(s,T_WSPACE); }

    BAR                  { RETURN(s,T_BAR); }

    DEC                  { RETURN(s,T_DECOR); }

    CHRD                 { RETURN(s,T_CHORD); }

    ANN                  { RETURN(s,T_ANNOTATION); }
    
    STR                  { RETURN(s,T_STRING); }
    
    BRK                  { RETURN(s,T_BREAK); }
    
    any                  { return -1; }  
  */
  
}
/*
    "[" L [:] (any\[\135])* "]"
                         { RETURN(s,T_INFIELD); }
*/

static int abcNT_NOTE(abcScanner *s)
{

  YYCTYPE *cursor, *limit, *marker;
  cursor = s->cursor;
  limit  = s->limit;
  marker = s->marker;
  /*!re2c
    
    SOL                  { RETURN(s,T_STARTLINE); }
    
    "[" L [:]            { s->cursor++;
                           s->nextstate= S_INFIELD | *(s->cursor);
                           RETURN(s,T_INFIELD); }

    WS+                  { RETURN(s,T_WSPACE); }

    N                    { RETURN(s,T_NOTE); }

    [z] (D | [/])*       { RETURN(s,T_REST); }

    [x] (D | [/])*       { RETURN(s,T_INVREST); }

    [Z] D*               { RETURN(s,T_MULTIREST); }

    [y]                  { RETURN(s,T_SPACER); }

    BAR                  { RETURN(s,T_BAR); }

    DEC                  { RETURN(s,T_DECOR); }

    [(] D ([:] D?)*      { RETURN(s,T_NPLET); }

    [(]                  { RETURN(s,T_OPENSLUR); }

    [)]                  { RETURN(s,T_CLOSESLUR); }

    [<]+                 { RETURN(s,T_DOTRIGHT); }

    [>]+                 { RETURN(s,T_DOTLEFT); }

    MIN                  { RETURN(s,T_TIE); }

    "["? D (D | "," | "-")*
                         { RETURN(s,T_REPEAT); }
    
    "["                  { RETURN(s,T_CHORDSTART); }
    
    "]"                  { RETURN(s,T_CHORDEND); }
    
    CHRD                 { RETURN(s,T_CHORD); }

    ANN                  { RETURN(s,T_ANNOTATION); }
    
    STR                  { RETURN(s,T_STRING); }

    "{/"                 { RETURN(s,T_GRACEACC); }
    [{]                  { RETURN(s,T_GRACEAPP); }
    [}]                  { RETURN(s,T_GRACEEND); }
    
    [\041]               { RETURN(s,T_BREAK); }
    
    any                  { return -1; }  
    */
}


/*
** This is the main function. It recognize a new token in the buffer,
** sets the scanner internals appropriately and returns the lenght
** of the token.
*/

int abcNextToken(abcScanner *s)
{

  YYCTYPE *cursor, *limit, *marker;
  int res=-1;

  /* Avoid calls past the end of file */  
  if (s->state == S_EOF) return -1;

  /* Reset the character saved if any */
  if (m_lastch(s)) *(s->cursor)=m_lastch(s);

  /* Change to a new state if required */
  if (s->nextstate != S_NONE) {
    s->state=s->nextstate;
    s->nextstate = S_NONE;
  }

  /* Get a new line */  
  if (m_lineempty(s)) abcGetLine(s);

  /* Line is still empty? Must be the end of file!! */  
  if (m_lineempty(s)) {
    m_eof(s);
    return -1;
  }

  cursor = s->cursor;
  limit  = s->limit;
  marker = s->marker;

  /* Check for Tokens valid regardless the state*/
  res=abcNT(s);
  if (res >= 0) return res;

  /* If we are in the "limbo" (before o between
  ** songs) consider any line as simple text.
  */
  if (s->state == S_LIMBO) {
    YYCURSOR=YYLIMIT;
    RETURN(s,T_TEXT);
  }

  /* Look for syllables in a lyrics line (w:) */
  if (s->state == (S_FIELD | 'w'))  res=abcNT_FIELD_w(s);
  if (res >= 0) return res;

  /* Look for symbols in a symbol line (s:) */
  if (s->state == (S_FIELD | 's'))  res=abcNT_FIELD_s(s);
  if (res >= 0) return res;

  /* Return the chars up to "]" as content for the fields */
  if (s->state & S_INFIELD) {
    while (*YYCURSOR && (*YYCURSOR != ']')) YYCURSOR++;
    if (*YYCURSOR == ']') {
      *YYCURSOR++='\0';
      s->nextstate=S_TUNE;
    }
    else {
      YYCURSOR--;
      while ((YYCURSOR > s->cursor) && isspace(*YYCURSOR))
        YYCURSOR--;
      if (*YYCURSOR != '\\') YYCURSOR++;
    }
    RETURN(s,T_FIELDB);
  }
  /**/
  /* Return the entire line as content for the field */
  if (s->state & (S_EXTFIELD | S_FIELD)) {
    YYCURSOR=YYLIMIT-1;
    while ((YYCURSOR > s->cursor) && isspace(*YYCURSOR))
      YYCURSOR--;
    if (*YYCURSOR != '\\') YYCURSOR++;
    RETURN(s,T_FIELDB);
  }

  /* Get notes and such from the tune */  
  if (s->state == S_TUNE) res=abcNT_NOTE(s);

  /* Nothing recognized! Report as "unkonwn" */  
  if (res == -1) {
    YYCURSOR=s->cursor+1;
    RETURN(s,T_UNKNOWN);
  }
  
  return res;
}

/*
** Scanner creation routines
*/

void abcClose(abcScanner *s)
{
   if (m_guarded(s)) {
     if (s->file)   fclose(s->file);
     if (s->buffer) free(s->buffer);
     
     s->buffer=NULL;
     s->file=NULL;
     m_unguard(s);
     free(s);
   }
}

abcScanner *abcNew(ULONG bufsize)
{
  abcScanner *s=NULL;
  ULONG realsize;

  s=malloc(sizeof(abcScanner));
  if (s) {
    m_unguard(s);
    if (bufsize < MINBUFSIZE) bufsize=MINBUFSIZE;
    if (bufsize > MAXBUFSIZE) bufsize=MAXBUFSIZE;
    s->bufsize=bufsize;
    realsize=s->bufsize + 16;
    s->buffer=malloc(realsize);
    if (s->buffer) {
      memset(s->buffer,0,realsize);
      s->line = 0;
      s->state = S_LIMBO;
      s->nextstate = S_NONE;
      s->token = T_NONE;
      s->file = NULL;
      m_guard(s);
      m_resetbuf(s);
      return (s);
    }
  }
  if (s) free(s);
  return (NULL);
}

abcScanner *abcFileScanner(char *filename,ULONG bufsize)
{
  abcScanner *s=NULL;
  
  if (filename != NULL) {
    s=abcNew(bufsize + strlen(filename)+1);
    if (s) {
      s->file=fopen(filename,"r");
      if (s->file) {
        s->bufsize -= strlen(filename)+1;
        s->filename=s->buffer+s->bufsize;
        strcpy(s->filename,filename);
      }
      else {
        abcClose(s);
        s=NULL;
      }
    }
  } 
  return(s);
}

abcScanner *abcStringScanner(char *tune)
{
  abcScanner *s=NULL;
  char *b;
  
  if (tune != NULL) {
    s=abcNew(strlen(tune) + 1 + 2);
    /*printf("+\t[%s]\n",tune);*/
    if (s) {
      b=s->buffer;
      *b++='\001';
      while (*tune && (*tune != '\n')) 
        *b++ = *tune++;
      *b++ = *tune++;
      if (*(b-1))
        *b++ = '\0';
      s->bufaux=b;
      s->limit=s->bufaux-1;
      while (*tune) 
        *b++ = *tune++;
      *b = '\0';
      s->line=1;
    }
    /*printf("-\t[%s]\n",s->buffer);
    printf("-\t[%s]\n",s->bufaux);*/
  } 
  return(s);
}

static int abcScan(abcScanner *as,abcHandler handler)
{
   if (as) {
    while (as && (as->state != S_EOF)) {
      abcNextToken(as);
      if (handler) {
        if (handler(as) > 0)
          m_eof(as);
      }
      else {
        printf("%s\t\t", abcTokenName(as->token));
        printf("[%s]\n", as->tokstr);
      }
    }
    abcClose(as);
    return 0;
  }
  return 1; 
}

int ABCAPI abcScanFile(char *filename,abcHandler handler)
{
  abcScanner *as;
    
  as=abcFileScanner(filename,1024);
  return(abcScan(as,handler));
}


int ABCAPI abcScanString(char *buffer,abcHandler handler)
{
  abcScanner *as;
    
  as=abcStringScanner(buffer);
  return(abcScan(as,handler));
}


const char * ABCAPI abcTokenName(Tokens n){
  return abcTokName[((n>T_FIRST && n<T_LAST)?n:T_UNKNOWN)];
}

const char * ABCAPI abcStateName(States s) {
  static char name[80];
  
  name[0]='\0';
  
  if (s == S_EOF)          return (abcStName[0]);
  else if (s == S_ERROR)   return (abcStName[1]);
  else if (s == S_LIMBO)   return (abcStName[2]);
  else if (s == S_NONE)    return (abcStName[3]);
  else if (s == S_TUNE)    return (abcStName[4]);
  else if (s & S_FIELD)    strcpy(name,abcStName[5]);
  else if (s & S_EXTFIELD) strcpy(name,abcStName[6]);
  else if (s & S_INFIELD)  strcpy(name,abcStName[7]);
  
  if ((s & (S_FIELD | S_INFIELD)) && (s & 0xFF)) {
    char *l=name;
    while (*l) l++;
    sprintf(l,"_%c",s & 0xFF);
  }
  return(name);
}

Tokens ABCAPI abcToken(abcScanner *s)
{ return(s->token); }

States ABCAPI abcState(abcScanner *s)
{ return(s->state); }

const char * ABCAPI abcString(abcScanner *s)
{ return(s->tokstr); }

const char * ABCAPI abcFilename(abcScanner *s)
{ return(s->filename); }

int ABCAPI abcLine(abcScanner *s)
{ return(s->line); }

int ABCAPI abcColumn(abcScanner *s)
{ return(m_column(s)); }


